From 419e2ee28d592472b038eb1bb5f2fa724c31a62f Mon Sep 17 00:00:00 2001
From: =?UTF-8?q?Hugo=20Beauz=C3=A9e-Luyssen?= <beauze.h@gmail.com>
Date: Sat, 18 Feb 2012 00:06:18 +0100
Subject: [PATCH 05/12] bluray: Adding support for overlay.

---
 modules/access/bluray.c |  436 ++++++++++++++++++++++++++++++++++++++++++++++-
 1 files changed, 427 insertions(+), 9 deletions(-)

diff --git a/modules/access/bluray.c b/modules/access/bluray.c
index 199f83f..ca5b110 100644
--- a/modules/access/bluray.c
+++ b/modules/access/bluray.c
@@ -36,9 +36,11 @@
 #include <vlc_demux.h>                      /* demux_t */
 #include <vlc_input.h>                      /* Seekpoints, chapters */
 #include <vlc_dialog.h>                     /* BD+/AACS warnings */
+#include <vlc_vout.h>                       /* vout_PutSubpicture / subpicture_t */
 
 #include <libbluray/bluray.h>
 #include <libbluray/meta_data.h>
+#include <libbluray/overlay.h>
 
 /*****************************************************************************
  * Module descriptor
@@ -66,6 +68,16 @@ vlc_module_begin ()
     set_callbacks( blurayOpen, blurayClose )
 vlc_module_end ()
 
+/* libbluray's overlay.h defines 2 types of overlay (bd_overlay_plane_e). */
+#define MAX_OVERLAY 2
+
+typedef enum    OverlayStatus
+{
+    Closed = 0,
+    ToDisplay,  //Used to mark the overlay to be displayed the first time.
+    Displayed,
+    Outdated    //used to update the overlay after it has been sent to the vout
+}               OverlayStatus;
 
 struct demux_sys_t
 {
@@ -76,10 +88,26 @@ struct demux_sys_t
     unsigned int    i_longest_title;
     input_title_t **pp_title;
 
+    /* Overlay stuff */
+    vlc_mutex_t             overlay_lock;
+    subpicture_t            *p_pic[MAX_OVERLAY];
+    OverlayStatus           overlay_status[MAX_OVERLAY];
+    subpicture_region_t     *p_regions[MAX_OVERLAY];
+    int             current_overlay; // -1 if no current overlay;
+
+    input_thread_t  *p_input;
+    vout_thread_t   *p_vout;
+
     /* TS stream */
     stream_t       *p_parser;
 };
 
+struct subpicture_updater_sys_t
+{
+    demux_sys_t *p_demux_sys;
+    int         i_plane;
+};
+
 /*****************************************************************************
  * Local prototypes
  *****************************************************************************/
@@ -90,6 +118,8 @@ static int     blurayDemuxMenu(demux_t *);
 static int     blurayInitTitles(demux_t *p_demux );
 static int     bluraySetTitle(demux_t *p_demux, int i_title);
 
+static void    blurayOverlayProc( void *ptr, const BD_OVERLAY * const overlay );
+
 #define FROM_TICKS(a) (a*CLOCK_FREQ / INT64_C(90000))
 #define TO_TICKS(a)   (a*INT64_C(90000)/CLOCK_FREQ)
 #define CUR_LENGTH    p_sys->pp_title[p_demux->info.i_title]->i_length
@@ -113,11 +143,12 @@ static int blurayOpen( vlc_object_t *object )
     }
 
     /* */
-    p_demux->p_sys = p_sys = malloc(sizeof(*p_sys));
+    p_demux->p_sys = p_sys = calloc( 1, sizeof(*p_sys) );
     if (unlikely(!p_sys)) {
         return VLC_ENOMEM;
     }
-    p_sys->p_parser = NULL;
+    vlc_mutex_init( &p_sys->overlay_lock );
+    p_sys->current_overlay = -1;
 
     /* init demux info fields */
     p_demux->info.i_update    = 0;
@@ -207,11 +238,16 @@ static int blurayOpen( vlc_object_t *object )
     bool b_menu = var_InheritBool( p_demux, "bluray-menu" );
     if ( b_menu )
     {
+        p_sys->p_input = demux_GetParentInput( p_demux );
+        if ( unlikely( p_sys->p_input == NULL ) )
+            goto error;
         /*
          * libbluray will start playback from "First-Title" title
          * Therefore, We don't have to select any title.
          */
         bd_play( p_sys->bluray );
+        /* Registering overlay event handler */
+        bd_register_overlay_proc( p_sys->bluray, p_demux, blurayOverlayProc );
     }
     else
     {
@@ -256,20 +292,392 @@ static void blurayClose( vlc_object_t *object )
     demux_t *p_demux = (demux_t*)object;
     demux_sys_t *p_sys = p_demux->p_sys;
 
+    /*
+     * Close libbluray first. This will close all the overlays before we release p_vout
+     *
+     * bd_close( NULL ) can crash
+     */
+    assert(p_sys->bluray);
+    bd_close(p_sys->bluray);
+
+    if ( p_sys->p_vout != NULL )
+        vlc_object_release( p_sys->p_vout );
+    if ( p_sys->p_input != NULL )
+        vlc_object_release( p_sys->p_input );
     if (p_sys->p_parser)
         stream_Delete(p_sys->p_parser);
-
+    vlc_mutex_destroy( &p_sys->overlay_lock );
     /* Titles */
     for (unsigned int i = 0; i < p_sys->i_title; i++)
         vlc_input_title_Delete(p_sys->pp_title[i]);
     TAB_CLEAN( p_sys->i_title, p_sys->pp_title );
 
-    /* bd_close( NULL ) can crash */
-    assert(p_sys->bluray);
-    bd_close(p_sys->bluray);
     free(p_sys);
 }
 
+/*****************************************************************************
+ * subpicture_updater_t functions:
+ *****************************************************************************/
+static int      subpictureUpdaterValidate( subpicture_t *p_subpic,
+                                           bool b_fmt_src, const video_format_t *p_fmt_src,
+                                           bool b_fmt_dst, const video_format_t *p_fmt_dst,
+                                           mtime_t i_ts )
+{
+    VLC_UNUSED( b_fmt_src ); VLC_UNUSED( b_fmt_dst );
+    VLC_UNUSED( p_fmt_src ); VLC_UNUSED( p_fmt_dst );
+    VLC_UNUSED( i_ts );
+    subpicture_updater_sys_t    *p_upd_sys = p_subpic->updater.p_sys;
+    demux_sys_t                 *p_sys = p_upd_sys->p_demux_sys;
+
+    vlc_mutex_lock( &p_sys->overlay_lock );
+    int res = p_sys->overlay_status[p_upd_sys->i_plane] == Outdated;
+    vlc_mutex_unlock( &p_sys->overlay_lock );
+    return res;
+}
+
+/* This should probably be moved to subpictures.c afterward */
+static subpicture_region_t*     subpicture_region_Clone( subpicture_region_t *p_region_src )
+{
+    if ( p_region_src == NULL )
+        return NULL;
+    subpicture_region_t *p_region_dst = subpicture_region_New( &p_region_src->fmt );
+    if ( unlikely( p_region_dst == NULL ) )
+        return NULL;
+    p_region_dst->i_x = p_region_src->i_x;
+    p_region_dst->i_y = p_region_src->i_y;
+    p_region_dst->i_align = p_region_src->i_align;
+    p_region_dst->i_alpha = p_region_src->i_alpha;
+
+    p_region_dst->psz_text = p_region_src->psz_text ? strdup( p_region_src->psz_text ) : NULL;
+    p_region_dst->psz_html = p_region_src->psz_html ? strdup( p_region_src->psz_html ) : NULL;
+    if ( p_region_src->p_style != NULL )
+    {
+        p_region_dst->p_style = malloc( sizeof( *p_region_dst->p_style ) );
+        p_region_dst->p_style = text_style_Copy( p_region_dst->p_style,
+                                                 p_region_src->p_style );
+    }
+
+    //Palette is already copied by subpicture_region_New, we just have to duplicate p_pixels
+    for ( int i = 0; i < p_region_src->p_picture->i_planes; i++ )
+        memcpy( p_region_dst->p_picture->p[i].p_pixels,
+                p_region_src->p_picture->p[i].p_pixels,
+                p_region_src->p_picture->p[i].i_lines * p_region_src->p_picture->p[i].i_pitch );
+    return p_region_dst;
+}
+
+static void     subpictureUpdaterUpdate( subpicture_t *p_subpic,
+                              const video_format_t *p_fmt_src,
+                              const video_format_t *p_fmt_dst,
+                              mtime_t i_ts )
+{
+    VLC_UNUSED( p_fmt_src ); VLC_UNUSED( p_fmt_dst ); VLC_UNUSED( i_ts );
+    subpicture_updater_sys_t    *p_upd_sys = p_subpic->updater.p_sys;
+    demux_sys_t                 *p_sys = p_upd_sys->p_demux_sys;
+
+    /*
+     * When this function is called, all p_subpic regions are gone.
+     * We need to duplicate our regions (stored internaly) to this subpic.
+     */
+    vlc_mutex_lock( &p_sys->overlay_lock );
+    p_subpic->p_region = subpicture_region_Clone( p_sys->p_regions[p_upd_sys->i_plane] );
+    if ( p_subpic->p_region == NULL )
+    {
+        vlc_mutex_unlock( &p_sys->overlay_lock );
+        return ;
+    }
+    subpicture_region_t *p_src = p_sys->p_regions[p_upd_sys->i_plane]->p_next;
+    subpicture_region_t *p_dst = p_subpic->p_region;
+    while ( p_src != NULL )
+    {
+        p_dst->p_next = subpicture_region_Clone( p_src );
+        if ( p_dst->p_next == NULL )
+            break ;
+        p_dst = p_dst->p_next;
+        p_src = p_src->p_next;
+    }
+    p_dst->p_next = NULL;
+    p_sys->overlay_status[p_upd_sys->i_plane] = Displayed;
+    vlc_mutex_unlock( &p_sys->overlay_lock );
+}
+
+static void     subpictureUpdaterDestroy( subpicture_t *p_subpic )
+{
+    free( p_subpic->updater.p_sys );
+}
+
+/*****************************************************************************
+ * libbluray overlay handling:
+ *****************************************************************************/
+static void blurayDeleteOverlaySubPic( demux_t *p_demux, int i_layer )
+{
+    demux_sys_t     *p_sys = p_demux->p_sys;
+
+    vlc_mutex_lock( &p_sys->overlay_lock );
+    p_sys->overlay_status[i_layer] = Closed;
+    if ( p_sys->p_pic[i_layer] == NULL )
+    {
+        vlc_mutex_unlock( &p_sys->overlay_lock );
+        return ;
+    }
+    //If we have a vout, let it cleanup by itself.
+    if ( p_sys->p_vout )
+        vout_FlushSubpictureChannel( p_sys->p_vout, p_sys->p_pic[i_layer]->i_channel );
+    else
+    {
+        /*
+         * Else, the subpicture hasn't been sent yet, clean it ourselves.
+         * In that case, the vout doesn't hold the picture, we can use it safely
+         */
+        subpicture_Delete( p_sys->p_pic[i_layer] );
+    }
+    p_sys->p_pic[i_layer] = NULL;
+    vlc_mutex_unlock( &p_sys->overlay_lock );
+}
+
+static void blurayCloseOverlay( demux_t *p_demux )
+{
+    demux_sys_t     *p_sys = p_demux->p_sys;
+
+    for ( int i = 0; i < MAX_OVERLAY; ++i )
+        blurayDeleteOverlaySubPic( p_demux, i );
+    p_demux->p_sys->current_overlay = -1;
+    if ( p_sys->p_vout != NULL )
+    {
+        vlc_object_release( p_sys->p_vout );
+        p_sys->p_vout = NULL;
+    }
+}
+
+/*
+ * Mark the overlay as "ToDisplay".
+ * This will not send the overlay to the vout instantly, as the vout
+ * may not be acquired (not acquirable) yet
+ * If is has already been acquire, the overlay has already been sent to it,
+ * therefore, we only flag the overlay as "Outdated"
+ */
+static void blurayActivateOverlay( demux_t *p_demux, const BD_OVERLAY* const ov )
+{
+    demux_sys_t     *p_sys = p_demux->p_sys;
+
+    /*
+     * If the overlay is already displayed, mark the picture as outdated.
+     * We must NOT use vout_PutSubpicture if a picture is already displayed.
+     */
+    vlc_mutex_lock( &p_sys->overlay_lock );
+    if ( ( p_sys->overlay_status[ov->plane] == Displayed ||
+           p_sys->overlay_status[ov->plane] == Outdated )
+         && p_sys->p_vout )
+    {
+        p_sys->overlay_status[ov->plane] = Outdated;
+        vlc_mutex_unlock( &p_sys->overlay_lock );
+        return ;
+    }
+    /*
+     * Mark the overlay as available, but don't display it right now.
+     * the blurayDemuxMenu will send it to vout, as it may be unavailable when
+     * the overlay is computed
+     */
+    p_sys->current_overlay = ov->plane;
+    p_sys->overlay_status[ov->plane] = ToDisplay;
+    vlc_mutex_unlock( &p_sys->overlay_lock );
+}
+
+static void blurayInitOverlay( demux_t *p_demux, const BD_OVERLAY* const ov )
+{
+    demux_sys_t     *p_sys = p_demux->p_sys;
+
+    if ( p_sys->p_pic[ov->plane] != NULL )
+        blurayDeleteOverlaySubPic( p_demux, ov->plane );
+
+    subpicture_updater_sys_t    *p_upd_sys = malloc( sizeof( *p_upd_sys ) );
+    if ( unlikely( p_upd_sys == NULL ) )
+        return ;
+    p_upd_sys->p_demux_sys = p_sys;
+    p_upd_sys->i_plane = ov->plane;
+    subpicture_updater_t    updater =
+    {
+        .pf_validate = subpictureUpdaterValidate,
+        .pf_update = subpictureUpdaterUpdate,
+        .pf_destroy = subpictureUpdaterDestroy,
+        .p_sys = p_upd_sys,
+    };
+    p_sys->p_pic[ov->plane] = subpicture_New( &updater );
+    p_sys->p_pic[ov->plane]->i_original_picture_width = ov->w;
+    p_sys->p_pic[ov->plane]->i_original_picture_height = ov->h;
+    p_sys->p_pic[ov->plane]->b_ephemer = true;
+    p_sys->p_pic[ov->plane]->b_absolute = true;
+}
+
+/**
+ * Destroy every regions in the subpictures.
+ * This is done in two steps:
+ * - Wiping our private regions list
+ * - Flagging the overlay as outdated, so the changes are replicated from
+ *   the subpicture_updater_t::pf_update
+ * This doesn't destroy the subpicture, as the overlay may be used again by libbluray.
+ */
+static void blurayClearOverlay( demux_t *p_demux, const BD_OVERLAY* const ov )
+{
+    demux_sys_t     *p_sys = p_demux->p_sys;
+
+    vlc_mutex_lock( &p_sys->overlay_lock );
+    subpicture_region_ChainDelete( p_sys->p_regions[ov->plane] );
+    p_sys->p_regions[ov->plane] = NULL;
+    p_sys->overlay_status[ov->plane] = Outdated;
+    vlc_mutex_unlock( &p_sys->overlay_lock );
+}
+
+/*
+ * This will draw to the overlay by adding a region to our region list
+ * This will have to be copied to the subpicture used to render the overlay.
+ */
+static void blurayDrawOverlay( demux_t *p_demux, const BD_OVERLAY* const ov )
+{
+    demux_sys_t     *p_sys = p_demux->p_sys;
+
+    /*
+     * Compute a subpicture_region_t.
+     * It will be copied and sent to the vout later.
+     */
+    if ( ov->img == NULL )
+        return ;
+    video_format_t          fmt;
+    video_format_Init( &fmt, 0 );
+    video_format_Setup( &fmt, VLC_CODEC_YUVP, ov->w, ov->h, 1, 1 );
+    vlc_mutex_lock( &p_sys->overlay_lock );
+
+    subpicture_region_t     *region = subpicture_region_New( &fmt );
+    region->i_x = ov->x;
+    region->i_y = ov->y;
+
+    const BD_PG_RLE_ELEM    *img = ov->img;
+    for ( int y = 0; y < ov->h; y++ )
+    {
+        for ( int x = 0; x < ov->w; )
+        {
+            memset( region->p_picture->p[0].p_pixels +
+                    y * region->p_picture->p[0].i_pitch + x,
+                    img->color, img->len );
+            x += img->len;
+            img++;
+        }
+    }
+    if ( ov->palette )
+    {
+        region->fmt.p_palette->i_entries = 256;
+        for ( int i = 0; i < 256; ++i )
+        {
+            region->fmt.p_palette->palette[i][0] = ov->palette[i].Y;
+            region->fmt.p_palette->palette[i][1] = ov->palette[i].Cb;
+            region->fmt.p_palette->palette[i][2] = ov->palette[i].Cr;
+            region->fmt.p_palette->palette[i][3] = ov->palette[i].T;
+        }
+    }
+    /*
+     * Append the new region at the end of our current regions:
+     * If there's no region yet, don't bother checking for update...
+     */
+    subpicture_region_t *p_reg = p_sys->p_regions[ov->plane];
+    if ( p_reg == NULL )
+    {
+        p_sys->p_regions[ov->plane] = region;
+        vlc_mutex_unlock( &p_sys->overlay_lock );
+        return ;
+    }
+    /*
+     * While adding the region, check if any region matches the one we're drawing.
+     * if so, delete the old region, and add the new one at the back of the list.
+     */
+    subpicture_region_t *p_prev = NULL;
+    while ( p_reg )
+    {
+        if ( p_reg->i_x == region->i_x && p_reg->i_y == region->i_y &&
+             p_reg->fmt.i_width == region->fmt.i_width &&
+             p_reg->fmt.i_height == region->fmt.i_height )
+        {
+            if ( p_prev == NULL )
+            {
+                p_sys->p_regions[ov->plane] = p_reg->p_next;
+                subpicture_region_Delete( p_reg );
+                p_reg = p_sys->p_regions[ov->plane];
+            }
+            else
+            {
+                p_prev->p_next = p_reg->p_next;
+                subpicture_region_Delete( p_reg );
+                p_reg = p_prev->p_next;
+            }
+        }
+        else
+        {
+            p_prev = p_reg;
+            p_reg = p_reg->p_next;
+        }
+    }
+    /*
+     * If there's only one region which has been updated,
+     * both prev and reg will be NULL
+     */
+    if ( p_prev == NULL )
+        p_sys->p_regions[ov->plane]->p_next = region;
+    else
+        p_prev->p_next = region;
+    vlc_mutex_unlock( &p_sys->overlay_lock );
+    /*
+     * /!\ The region is now stored in our internal list, but not in the subpicture /!\
+     */
+}
+
+static void blurayOverlayProc( void *ptr, const BD_OVERLAY *const overlay )
+{
+    demux_t     *p_demux = (demux_t*)ptr;
+
+    if ( overlay == NULL )
+    {
+        msg_Info( p_demux, "Closing overlay." );
+        blurayCloseOverlay( p_demux );
+        return ;
+    }
+    switch ( overlay->cmd )
+    {
+        case BD_OVERLAY_INIT:
+            msg_Info( p_demux, "Initializing overlay" );
+            blurayInitOverlay( p_demux, overlay );
+            break ;
+        case BD_OVERLAY_CLEAR:
+            blurayClearOverlay( p_demux, overlay );
+            break ;
+        case BD_OVERLAY_FLUSH:
+            blurayActivateOverlay( p_demux, overlay );
+            break ;
+        case BD_OVERLAY_DRAW:
+            blurayDrawOverlay( p_demux, overlay );
+            break ;
+        default:
+            msg_Warn( p_demux, "Unknown BD overlay command: %u", overlay->cmd );
+            break ;
+    }
+}
+
+static void bluraySendOverlayToVout( demux_t *p_demux )
+{
+    demux_sys_t *p_sys = p_demux->p_sys;
+
+    assert( p_sys->current_overlay >= 0 &&
+            p_sys->p_pic[p_sys->current_overlay] != NULL );
+
+    p_sys->p_pic[p_sys->current_overlay]->i_start =
+            p_sys->p_pic[p_sys->current_overlay]->i_stop = mdate();
+    p_sys->p_pic[p_sys->current_overlay]->i_channel =
+            vout_RegisterSubpictureChannel( p_sys->p_vout );
+    /*
+     * After this point, the picture should not be accessed from the demux thread,
+     * as it's hold by the vout thread.
+     */
+    vout_PutSubpicture( p_sys->p_vout, p_sys->p_pic[p_sys->current_overlay] );
+    /* Mark the picture as Outdated, as it contains no region for now. */
+    p_sys->overlay_status[p_sys->current_overlay] = Outdated;
+}
 
 static int blurayInitTitles(demux_t *p_demux )
 {
@@ -570,9 +978,7 @@ static int blurayDemuxMenu( demux_t *p_demux )
      */
     if ( nread == 0 )
     {
-        if ( e.event == BD_EVENT_NONE )
-            msg_Info( p_demux, "We reached the end of a title" );
-        else
+        if ( e.event != BD_EVENT_NONE )
             blurayHandleEvent( p_demux, &e );
         block_Release(p_block);
         return 1;
@@ -584,6 +990,18 @@ static int blurayDemuxMenu( demux_t *p_demux )
     }
     p_block->i_buffer = nread;
 
+    vlc_mutex_lock( &p_sys->overlay_lock );
+    if ( p_sys->current_overlay != -1 &&
+         p_sys->overlay_status[p_sys->current_overlay] == ToDisplay )
+    {
+        vlc_mutex_unlock( &p_sys->overlay_lock );
+        if ( p_sys->p_vout == NULL )
+            p_sys->p_vout = input_GetVout( p_sys->p_input );
+        if ( p_sys->p_vout != NULL )
+            bluraySendOverlayToVout( p_demux );
+    }
+    else
+        vlc_mutex_unlock( &p_sys->overlay_lock );
     stream_DemuxSend( p_sys->p_parser, p_block );
     return 1;
 }
-- 
1.7.9.1

